==> pyrefly-0.13.1.tar.gz: pyrefly-0.13.1/pyrefly/lib/commands/suppress.rs <==
/*
* Copyright (c) Meta Platforms, Inc. and affiliates.
*
* This source code is licensed under the MIT license found in the
* LICENSE file in the root directory of this source tree.
*/

use std::path::PathBuf;

use regex::Regex;
use ruff_source_file::OneIndexed;
use starlark_map::small_map::SmallMap;
use starlark_map::small_set::SmallSet;
use tracing::error;

use crate::error::error::Error;
use crate::util::fs_anyhow;

/// Combines all errors that affect one line into a single entry.
// The current format is: `# pyrefly: ignore  # error1, error2, ...`, because pyrefly does not currently support pyre-fixme.
fn dedup_errors(errors: &[Error]) -> SmallMap<usize, String> {
    let mut deduped_errors = SmallMap::new();
    for error in errors {
        let e: &mut String = deduped_errors
            .entry(error.source_range().start.row.to_zero_indexed())
            .or_default();
        if e.is_empty() {
            e.push_str("# pyrefly: ignore  # ");
        } else {
            e.push_str(", ");
        }
        e.push_str(error.error_kind().to_name());
    }
    deduped_errors
}

/// Adds error suppressions for the given errors in the given files.
/// Returns a list of files that failed to be be patched, and a list of files that were patched.
/// The list of failures includes the error that occurred, which may be a read or write error.
fn add_suppressions(
    path_errors: &SmallMap<PathBuf, Vec<Error>>,
) -> (Vec<(&PathBuf, anyhow::Error)>, Vec<&PathBuf>) {
    let mut failures = vec![];
    let mut successes = vec![];
    for (path, errors) in path_errors {
        let file = match fs_anyhow::read_to_string(path) {
            Ok(f) => f,
            Err(e) => {
                failures.push((path, e));
                continue;
            }
        };
        let deduped_errors = dedup_errors(errors);
        let mut buf = String::new();
        for (idx, line) in file.lines().enumerate() {
            if let Some(error_comment) = deduped_errors.get(&idx) {
                // As a simple formatting step, indent the error comment to match the line below it.
                if let Some(first_char) = line.find(|c: char| !c.is_whitespace()) {
                    buf.push_str(&line[..first_char]);
                }
                buf.push_str(error_comment);
                buf.push('\n');
            }
            buf.push_str(line);
            buf.push('\n');
        }
        if let Err(e) = fs_anyhow::write(path, buf.as_bytes()) {
            failures.push((path, e));
        } else {
            successes.push(path);
        }
    }
    (failures, successes)
}

pub fn suppress_errors(path_errors: &SmallMap<PathBuf, Vec<Error>>) {
    eprintln!("Inserting error suppressions...");
    if path_errors.is_empty() {
        eprintln!("No errors to suppress!");
        return;
    }
    let (failures, successes) = add_suppressions(path_errors);
    eprintln!(
        "Finished suppressing errors in {}/{} files",
        successes.len(),
        path_errors.len()
    );
    if !failures.is_empty() {
        eprintln!("Failed to suppress errors in {} files:", failures.len());
        for (path, e) in failures {
            eprintln!("  {path:#?}: {e}");
        }
    }
}

pub fn find_unused_ignores<'a>(
    all_ignores: SmallMap<&'a PathBuf, SmallSet<OneIndexed>>,
    suppressed_errors: SmallMap<&PathBuf, SmallSet<OneIndexed>>,
) -> SmallMap<&'a PathBuf, SmallSet<OneIndexed>> {
    let mut all_unused_ignores: SmallMap<&PathBuf, SmallSet<OneIndexed>> = SmallMap::new();
    let one = OneIndexed::new(1).unwrap();
    let default_set: SmallSet<OneIndexed> = SmallSet::new();
    // Loop over each path only save the ignores that are not in use
    for (path, ignores) in all_ignores {
        let errors = suppressed_errors.get(path).unwrap_or(&default_set);
        let mut unused_ignores = SmallSet::new();
        for ignore in ignores {
            if let Some(location) = ignore.checked_add(one)
                && !errors.contains(&location)
                && !errors.contains(&ignore)
            {
                unused_ignores.insert(ignore);
            }
        }

        all_unused_ignores.insert(path, unused_ignores);
    }
    all_unused_ignores
}

pub fn remove_unused_ignores(path_ignores: SmallMap<&PathBuf, SmallSet<OneIndexed>>) {
    let regex = Regex::new(r"# pyrefly: ignore.*$").unwrap();
    for (path, ignores) in path_ignores {
        let zero_index_ignores: SmallSet<usize> =
            ignores.iter().map(|i| i.to_zero_indexed()).collect();
        if let Ok(file) = fs_anyhow::read_to_string(path) {
            let mut buf = String::with_capacity(file.len());
            let lines = file.lines();
            for (idx, line) in lines.enumerate() {
                if zero_index_ignores.contains(&idx) {
                    // TODO: Expand support of what we remove and thoroughly test
                    let new_string = regex.replace_all(line, "");
                    if !new_string.trim().is_empty() {
                        buf.push_str(&new_string);
                    }
                    buf.push('\n');
                    continue;
                }
                buf.push_str(line);
                buf.push('\n');
            }
            if let Err(e) = fs_anyhow::write(path, buf.as_bytes()) {
                error!("Failed to remove unused error suppressions in {} files:", e);
            }
        }
    }
}

#[cfg(test)]
mod tests {
    use pretty_assertions::assert_str_eq;
    use ruff_source_file::OneIndexed;
    use ruff_source_file::SourceLocation;
    use tempfile;
    use vec1::Vec1;

    use super::*;
    use crate::error::kind::ErrorKind;
    use crate::module::module_info::SourceRange;
    use crate::module::module_path::ModulePath;

    fn sourcerange(row: usize, column: usize) -> SourceRange {
        let row = OneIndexed::new(row).unwrap();
        let column = OneIndexed::new(column).unwrap();
        SourceRange {
            start: SourceLocation { row, column },
            end: SourceLocation { row, column },
        }
    }

    fn error(path: PathBuf, row: usize, column: usize, error_kind: ErrorKind) -> Error {
        Error::new(
            ModulePath::filesystem(path),
            sourcerange(row, column),
            Vec1::new("test message".to_owned()),
            false,
            error_kind,
        )
    }

    fn test_remove_suppressions(lines: SmallSet<OneIndexed>, input: &str, want: &str) {
        let tdir = tempfile::tempdir().unwrap();
        let path = tdir.path().join("test.py");
        fs_anyhow::write(&path, input.as_bytes()).unwrap();
        let map = SmallMap::from_iter([(&path, lines)]);
        remove_unused_ignores(map);
        let got_file = fs_anyhow::read_to_string(&path).unwrap();
        assert_str_eq!(want, got_file);
    }

    #[test]
    fn test_add_suppressions() {
        let tdir = tempfile::tempdir().unwrap();
        let path = tdir.path().join("test.py");
        fs_anyhow::write(
            &path,
            br#"x: str = 1


def f(y: int) -> None:
    """Doc comment"""
    x = "one" + y
    return x


f(x)

"#,
        )
        .unwrap();
        // These errors were determined by manually running pyrefly on the snippet above.
        // As such, they may not be 100% accurate with what pyrefly produces now.
        let errors = {
            let mut e = SmallMap::new();
            e.insert(
                path.clone(),
                vec![
                    error(path.clone(), 1, 10, ErrorKind::BadAssignment),
                    error(path.clone(), 6, 7, ErrorKind::BadArgumentType),
                    error(path.clone(), 7, 10, ErrorKind::BadReturn),
                    error(path.clone(), 10, 3, ErrorKind::BadArgumentType),
                ],
            );
            e
        };

        let (got_failures, got_successes) = add_suppressions(&errors);
        assert!(got_failures.is_empty());
        assert_eq!(vec![&path], got_successes);

        let want_file = r#"# pyrefly: ignore  # bad-assignment
x: str = 1


def f(y: int) -> None:
    """Doc comment"""
    # pyrefly: ignore  # bad-argument-type
    x = "one" + y
    # pyrefly: ignore  # bad-return
    return x


# pyrefly: ignore  # bad-argument-type
f(x)

"#;
        let got_file = fs_anyhow::read_to_string(&path).unwrap();
        assert_str_eq!(want_file, got_file);
    }

    #[test]
    fn test_remove_suppression_above() {
        let lines = SmallSet::from_iter([OneIndexed::new(3).unwrap()]);
        let input = r#"
def f() -> int:
    # pyrefly: ignore # bad-return
    return 1
"#;
        let want = r#"
def f() -> int:

    return 1
"#;
        test_remove_suppressions(lines, input, want);
    }

    #[test]
    fn test_remove_suppression_above_two() {
        let lines = SmallSet::from_iter([OneIndexed::new(3).unwrap()]);
        let input = r#"
def g() -> str:
    # pyrefly: ignore # bad-return
    return "hello"
"#;
        let want = r#"
def g() -> str:

    return "hello"
"#;
        test_remove_suppressions(lines, input, want);
    }

    #[test]
    fn test_remove_suppression_inline() {
        let lines = SmallSet::from_iter([OneIndexed::new(3).unwrap()]);
        let input = r#"
def g() -> str:
    return "hello" # pyrefly: ignore # bad-return
"#;
        let want = r#"
def g() -> str:
    return "hello" 
"#;
        test_remove_suppressions(lines, input, want);
    }

    #[test]
    fn test_remove_suppression_multiple() {
        let lines = SmallSet::from_iter([OneIndexed::new(3).unwrap(), OneIndexed::new(5).unwrap()]);
        let input = r#"
def g() -> str:
    return "hello" # pyrefly: ignore # bad-return
def f() -> int:
    # pyrefly: ignore
    return 1
"#;
        let output = r#"
def g() -> str:
    return "hello" 
def f() -> int:

    return 1
"#;
        test_remove_suppressions(lines, input, output);
    }
}

==> pyrefly-0.15.1.tar.gz: pyrefly-0.15.1/pyrefly/lib/commands/suppress.rs <==
/*
* Copyright (c) Meta Platforms, Inc. and affiliates.
*
* This source code is licensed under the MIT license found in the
* LICENSE file in the root directory of this source tree.
*/

use std::path::PathBuf;

use regex::Regex;
use ruff_source_file::OneIndexed;
use starlark_map::small_map::SmallMap;
use starlark_map::small_set::SmallSet;
use tracing::error;

use crate::error::error::Error;
use crate::module::module_info::GENERATED_TOKEN;
use crate::ruff::ast::Ast;
use crate::util::fs_anyhow;

/// Combines all errors that affect one line into a single entry.
// The current format is: `# pyrefly: ignore  # error1, error2, ...`, because pyrefly does not currently support pyre-fixme.
fn dedup_errors(errors: &[Error]) -> SmallMap<usize, String> {
    let mut deduped_errors = SmallMap::new();
    for error in errors {
        let e: &mut String = deduped_errors
            .entry(error.source_range().start.row.to_zero_indexed())
            .or_default();
        let contains_error = e.contains(error.error_kind().to_name());
        if e.is_empty() {
            e.push_str("# pyrefly: ignore  # ");
        } else if !contains_error {
            e.push_str(", ");
        }

        if !contains_error {
            e.push_str(error.error_kind().to_name());
        }
    }
    deduped_errors
}

/// Adds error suppressions for the given errors in the given files.
/// Returns a list of files that failed to be be patched, and a list of files that were patched.
/// The list of failures includes the error that occurred, which may be a read or write error.
fn add_suppressions(
    path_errors: &SmallMap<PathBuf, Vec<Error>>,
) -> (Vec<(&PathBuf, anyhow::Error)>, Vec<&PathBuf>) {
    let mut failures = vec![];
    let mut successes = vec![];
    for (path, errors) in path_errors {
        let file = match fs_anyhow::read_to_string(path) {
            Ok(f) => f,
            Err(e) => {
                failures.push((path, e));
                continue;
            }
        };
        // Avoid adding suppressions to files that are not parsable.
        // Save AST for comparison later
        let (_ast, parse_errors) = Ast::parse(&file);
        if !parse_errors.is_empty() {
            eprintln!(
                "Unable to silence errors in `{}` because it is not parsable",
                path.display()
            );
            failures.push((path, anyhow::Error::msg("File is not parsable")));
            continue;
        }
        if file.contains(GENERATED_TOKEN) {
            eprintln!("Skipping `{}` because it is generated", path.display());
            failures.push((path, anyhow::Error::msg("Generated file")));
            continue;
        }
        let deduped_errors = dedup_errors(errors);
        let mut buf = String::new();
        for (idx, line) in file.lines().enumerate() {
            if let Some(error_comment) = deduped_errors.get(&idx) {
                // As a simple formatting step, indent the error comment to match the line below it.
                if let Some(first_char) = line.find(|c: char| !c.is_whitespace()) {
                    buf.push_str(&line[..first_char]);
                }
                buf.push_str(error_comment);
                buf.push('\n');
            }
            buf.push_str(line);
            buf.push('\n');
        }
        if let Err(e) = fs_anyhow::write(path, buf.as_bytes()) {
            failures.push((path, e));
        } else {
            successes.push(path);
        }
    }
    (failures, successes)
}

pub fn suppress_errors(path_errors: &SmallMap<PathBuf, Vec<Error>>) {
    eprintln!("Inserting error suppressions...");
    if path_errors.is_empty() {
        eprintln!("No errors to suppress!");
        return;
    }
    let (failures, successes) = add_suppressions(path_errors);
    eprintln!(
        "Finished suppressing errors in {}/{} files",
        successes.len(),
        path_errors.len()
    );
    if !failures.is_empty() {
        eprintln!("Failed to suppress errors in {} files:", failures.len());
        for (path, e) in failures {
            eprintln!("  {path:#?}: {e}");
        }
    }
}

pub fn find_unused_ignores<'a>(
    all_ignores: SmallMap<&'a PathBuf, SmallSet<OneIndexed>>,
    suppressed_errors: SmallMap<&PathBuf, SmallSet<OneIndexed>>,
) -> SmallMap<&'a PathBuf, SmallSet<OneIndexed>> {
    let mut all_unused_ignores: SmallMap<&PathBuf, SmallSet<OneIndexed>> = SmallMap::new();
    let one = OneIndexed::new(1).unwrap();
    let default_set: SmallSet<OneIndexed> = SmallSet::new();
    // Loop over each path only save the ignores that are not in use
    for (path, ignores) in all_ignores {
        let errors = suppressed_errors.get(path).unwrap_or(&default_set);
        let mut unused_ignores = SmallSet::new();
        for ignore in ignores {
            if let Some(location) = ignore.checked_add(one)
                && !errors.contains(&location)
                && !errors.contains(&ignore)
            {
                unused_ignores.insert(ignore);
            }
        }

        all_unused_ignores.insert(path, unused_ignores);
    }
    all_unused_ignores
}

pub fn remove_unused_ignores(path_ignores: SmallMap<&PathBuf, SmallSet<OneIndexed>>) {
    let regex = Regex::new(r"# pyrefly: ignore.*$").unwrap();
    for (path, ignores) in path_ignores {
        let zero_index_ignores: SmallSet<usize> =
            ignores.iter().map(|i| i.to_zero_indexed()).collect();
        if let Ok(file) = fs_anyhow::read_to_string(path) {
            let mut buf = String::with_capacity(file.len());
            let lines = file.lines();
            for (idx, line) in lines.enumerate() {
                if zero_index_ignores.contains(&idx) {
                    // TODO: Expand support of what we remove and thoroughly test
                    let new_string = regex.replace_all(line, "");
                    if !new_string.trim().is_empty() {
                        buf.push_str(new_string.trim_end());
                    }
                    buf.push('\n');
                    continue;
                }
                buf.push_str(line);
                buf.push('\n');
            }
            if let Err(e) = fs_anyhow::write(path, buf.as_bytes()) {
                error!("Failed to remove unused error suppressions in {} files:", e);
            }
        }
    }
}

#[cfg(test)]
mod tests {
    use pretty_assertions::assert_str_eq;
    use ruff_source_file::OneIndexed;
    use ruff_source_file::SourceLocation;
    use tempfile;
    use vec1::Vec1;

    use super::*;
    use crate::error::kind::ErrorKind;
    use crate::module::module_info::SourceRange;
    use crate::module::module_path::ModulePath;

    fn sourcerange(row: usize, column: usize) -> SourceRange {
        let row = OneIndexed::new(row).unwrap();
        let column = OneIndexed::new(column).unwrap();
        SourceRange {
            start: SourceLocation { row, column },
            end: SourceLocation { row, column },
        }
    }

    fn error(path: PathBuf, row: usize, column: usize, error_kind: ErrorKind) -> Error {
        Error::new(
            ModulePath::filesystem(path),
            sourcerange(row, column),
            Vec1::new("test message".to_owned()),
            false,
            error_kind,
        )
    }

    fn test_remove_suppressions(lines: SmallSet<OneIndexed>, input: &str, want: &str) {
        let tdir = tempfile::tempdir().unwrap();
        let path = tdir.path().join("test.py");
        fs_anyhow::write(&path, input.as_bytes()).unwrap();
        let map = SmallMap::from_iter([(&path, lines)]);
        remove_unused_ignores(map);
        let got_file = fs_anyhow::read_to_string(&path).unwrap();
        assert_str_eq!(want, got_file);
    }

    fn assert_suppress_errors(
        path_errors: Vec<(usize, usize, ErrorKind)>,
        input: &str,
        output: &str,
    ) {
        let tdir = tempfile::tempdir().unwrap();
        let path = tdir.path().join("test.py");
        fs_anyhow::write(&path, input.as_bytes()).unwrap();
        let errors = {
            let mut e = SmallMap::new();
            e.insert(
                path.clone(),
                path_errors
                    .into_iter()
                    .map(|x| error(path.clone(), x.0, x.1, x.2))
                    .collect(),
            );
            e
        };
        let (failures, successes) = add_suppressions(&errors);
        assert!(failures.is_empty());
        assert_eq!(vec![&path], successes);
        let got_file = fs_anyhow::read_to_string(&path).unwrap();
        assert_str_eq!(output, got_file);
    }

    fn assert_no_error_suppression(
        path_errors: Vec<(usize, usize, ErrorKind)>,
        input: &str,
        output: &str,
    ) {
        let tdir = tempfile::tempdir().unwrap();
        let path = tdir.path().join("test.py");
        fs_anyhow::write(&path, input.as_bytes()).unwrap();
        let errors = {
            let mut e = SmallMap::new();
            e.insert(
                path.clone(),
                path_errors
                    .into_iter()
                    .map(|x| error(path.clone(), x.0, x.1, x.2))
                    .collect(),
            );
            e
        };
        let (failures, _successes) = add_suppressions(&errors);
        assert!(!failures.is_empty());
        let got_file = fs_anyhow::read_to_string(&path).unwrap();
        assert_str_eq!(output, got_file);
    }

    #[test]
    fn test_add_suppressions() {
        assert_suppress_errors(
            vec![
                (1, 10, ErrorKind::BadAssignment),
                (6, 7, ErrorKind::BadArgumentType),
                (7, 10, ErrorKind::BadReturn),
                (10, 3, ErrorKind::BadArgumentType),
            ],
            r#"x: str = 1


def f(y: int) -> None:
    """Doc comment"""
    x = "one" + y
    return x


f(x)

"#,
            r#"# pyrefly: ignore  # bad-assignment
x: str = 1


def f(y: int) -> None:
    """Doc comment"""
    # pyrefly: ignore  # bad-argument-type
    x = "one" + y
    # pyrefly: ignore  # bad-return
    return x


# pyrefly: ignore  # bad-argument-type
f(x)

"#,
        );
    }
    #[test]
    fn test_add_suppressions_existing_comment() {
        assert_suppress_errors(
            vec![(3, 10, ErrorKind::BadAssignment)],
            r#"
# comment
def foo() -> None: pass
"#,
            r#"
# comment
# pyrefly: ignore  # bad-assignment
def foo() -> None: pass
"#,
        );
    }

    #[test]
    fn test_add_suppressions_duplicate_errors() {
        assert_suppress_errors(
            vec![
                (3, 10, ErrorKind::BadAssignment),
                (3, 10, ErrorKind::BadAssignment),
            ],
            r#"
# comment
def foo() -> None: pass
"#,
            r#"
# comment
# pyrefly: ignore  # bad-assignment
def foo() -> None: pass
"#,
        );
    }

    #[test]
    fn test_add_suppressions_multiple_errors_one_line() {
        assert_suppress_errors(
            vec![
                (3, 10, ErrorKind::BadAssignment),
                (3, 10, ErrorKind::TypeAliasError),
            ],
            r#"
# comment
def foo() -> None: pass
"#,
            r#"
# comment
# pyrefly: ignore  # bad-assignment, type-alias-error
def foo() -> None: pass
"#,
        );
    }

    #[test]
    fn test_add_suppressions_unparseable_line_break() {
        assert_no_error_suppression(
            vec![(3, 10, ErrorKind::BadAssignment)],
            r#"
def foo() -> None:
    line_break = \\
        [
            param
        ]
    unrelated_line = 0
        "#,
            r#"
def foo() -> None:
    line_break = \\
        [
            param
        ]
    unrelated_line = 0
        "#,
        );
    }

    #[test]
    fn test_no_suppress_generated_files() {
        let file_contents = format!(
            r#"
{}

def bar() -> None: 
pass 
    "#,
            GENERATED_TOKEN
        );
        assert_no_error_suppression(
            vec![(1, 10, ErrorKind::BadAssignment)],
            &file_contents,
            &file_contents,
        );
    }

    #[test]
    fn test_remove_suppression_above() {
        let lines = SmallSet::from_iter([OneIndexed::new(3).unwrap()]);
        let input = r#"
def f() -> int:
    # pyrefly: ignore # bad-return
    return 1
"#;
        let want = r#"
def f() -> int:

    return 1
"#;
        test_remove_suppressions(lines, input, want);
    }

    #[test]
    fn test_remove_suppression_above_two() {
        let lines = SmallSet::from_iter([OneIndexed::new(3).unwrap()]);
        let input = r#"
def g() -> str:
    # pyrefly: ignore # bad-return
    return "hello"
"#;
        let want = r#"
def g() -> str:

    return "hello"
"#;
        test_remove_suppressions(lines, input, want);
    }

    #[test]
    fn test_remove_suppression_inline() {
        let lines = SmallSet::from_iter([OneIndexed::new(3).unwrap()]);
        let input = r#"
def g() -> str:
    return "hello" # pyrefly: ignore # bad-return
"#;
        let want = r#"
def g() -> str:
    return "hello"
"#;
        test_remove_suppressions(lines, input, want);
    }

    #[test]
    fn test_remove_suppression_multiple() {
        let lines = SmallSet::from_iter([OneIndexed::new(3).unwrap(), OneIndexed::new(5).unwrap()]);
        let input = r#"
def g() -> str:
    return "hello" # pyrefly: ignore # bad-return
def f() -> int:
    # pyrefly: ignore
    return 1
"#;
        let output = r##"
def g() -> str:
    return "hello"
def f() -> int:

    return 1
"##;
        test_remove_suppressions(lines, input, output);
    }
}
